#include <list>
#include <set>
#include <vector>
#include "common.hh"
#include "RENode.hh"
#include "SRNode.hh"
#include "RETree.hh"
#include "debug.hh"
#include "Token.hh"
#include "Symbol.hh"

//
// Tree
//

Tree::Tree(RENode * r)
  : root(r)
{
  leaves = new RENodeList();
  leaves->push_back(root);
  debug(15) << "Tree::Tree: " << toString() << "\n";
}

/**
 * This is a shallow destructor only.
 * Use Tree::clear() to delete all nodes. 
 * clear will call this destructor for each node.
 */
Tree::~Tree() { delete leaves; }

RENodeI      Tree::begin() const    { return leaves->begin(); }
RENodeI      Tree::end() const      { return leaves->end(); }
unsigned int Tree::size() const     { return leaves->size(); }
TokenString* Tree::getInput() const { return inputString; }

#ifdef _OLD_STUFF_

/** O(1) 
 * Adds a branch to the specified node.
 *@return iterator pointing to the added child node.
 *@param parent The parent node of the new branch. 
 *@param rw The Symbol that matched (and rewrites) the Token after the parent node.
 */
RENodeI Tree::addBranch(RENodeI parent, Symbol* rw = 0, Token* exactToken = 0)
{
  RENode *p = *parent;
  RENodeI ci = parent;
  
  debug(8) << toString() << "->addBranch(" 
	   << p->toString() << ", " << (rw?rw->toString():0) << ")\n";
  
  // create the new child node.
  SRNode * next = p->getNext()->getNext(); 
  RENode * childNode = new RENode(next, p, rw, exactToken);
  
  // update the parent
  p->addChild();
  p->removeCopy();
  
  // update the leaf list
  leaves->insert(parent, childNode); // insert the child before the parent.
  --ci;                              // ci now points to the child.
  leaves->erase(parent);             // erase the parent from the leaf list.
  return ci;                         // the child is now where the parent
                                     // was before in the leaf list.
}

/** O(n); n = #of added branches
 * Adds a number of branches to the specified node.
 *@return iterator pointing to the lastly added child node.
 *@param parent The parent node of the new branch. 
 *@param rw The Symbol that matched (and rewrites) the Token after the parent node.
 *@param exactTokens vector of the exact tokens that were matched by the rewriter.
 */
RENodeI Tree::addBranch(RENodeI parent, Symbol * rw, vector<Token*>* exactTokens)
{
  RENode *p = *parent;
  RENodeI ci = parent;
  SRNode * next = p->getNext()->getNext(); 
  
  debug(8) << toString() << "->addBranch(" 
	   << p->toString() << ", " << (rw?rw->toString():0) << ")\n";
  
  for(vector<Token*>::iterator 
	i = exactTokens->begin(),
	z = exactTokens->end();
      i!=z;
      ++i)
    {
      // create the new child node.
      RENode * childNode = new RENode(next, p, rw, *i);
      
      // update the parent
      p->addChild();
      
      // update the leaf list
      leaves->insert(parent, childNode); // insert the child before the parent.
    }
  
  --ci;                  // ci now points to the child.
  leaves->erase(parent); // erase the parent from the leaf list.
  p->removeCopy();       // decrease the reference count of the parent.
  return ci;             // the last inserted child is now where the parent
                         // was before in the leaf list.
}

#endif

/** O(n); n = #of added branches
 * Adds a number of branches to the specified node.
 *@return iterator pointing to the lastly added child node.
 *@param parent The parent node of the new branch. 
 *@param rw The Symbol that matched (and rewrites) the Token after the parent node.
 *@param exactNodes vector of the exact nodes that must be added to the parent node.
 */
RENodeI Tree::addBranch(RENodeI parent, Symbol * rw, vector<RENode*>* exactNodes)
{
  RENode *p = *parent;
  RENodeI ci = parent;
  
  debug(8) << toString() << "->addBranch(" 
	   << p->toString() << ", " << (rw?rw->toString():0) << ")\n";
  
  for(vector<RENode*>::iterator 
	i = exactNodes->begin(),
	z = exactNodes->end();
      i!=z;
      ++i)
    {
      RENode * childNode = *i;
      
      // update the parent
      p->addChild();
      
      // update the leaf list
      leaves->insert(parent, childNode); // insert the child before the parent.
    }
  
  --ci;                  // ci now points to the child.
  leaves->erase(parent); // erase the parent from the leaf list.
  p->removeCopy();       // decrease the reference count of the parent.
  return ci;             // the last inserted child is now where the parent
                         // was before in the leaf list.
}

/** O(1) -> O(n); n = #nodes in tree 
 * @return The next node in the leaf list
 */
RENodeI Tree::removeBranch(RENodeI leaf)
{
  RENode *p, *n = *leaf;
  RENodeI ni = leaf;
  ++ni; // ni now points to the next leaf in the list.
  
  debug(2) << "*remove* " << n->toString() << "\n";
  
  // remove/dereference
  leaves->erase(leaf);
  n->removeCopy();
  
  // destroy dead part of the branch.
  while(n->isDead())
    {
      p = n->getParent();
      delete n;         // destroy node.
      if(!p) break;     // if n was the root, then there is nothing more to do.
      p->removeChild(); // remove the node from the parents childcount
      n = p;            // operate on parent in the next iteration.
    }
  
  return ni;
}

/** O(n); n = #nodes in tree 
 * Remove and destroy all nodes in tree.
 */
void Tree::clear()
{
  for(RENodeI i = leaves->begin(), z = leaves->end();
      i != z; i++)
    {
      i = removeBranch(i);
    }
}

//
// temporary work copies of trees:
//

/** O(n) */
Tree::Tree(Tree * tree) 
  : original(tree)
{
  leaves = new RENodeList();
  inputString = original->inputString;
  for(RENodeI i = tree->begin(), z = tree->end();
      i != z; i++)
    {
      RENode * node = *i;
      
      // increase each leaf's copycount
      // and add it to the leaf list.
      leaves->push_back(node);
      node->addCopy();
    }
}

/** O(n)
 * Commit the changes made in the copy to the original.
 */
void Tree::commit()
{
  // go through the original's leaf nodes
  RENodeI j, i = original->begin(), z = original->end();
  while(i != z)
    {
      j = i; 
      ++j;
      RENode * node = *i;
      
      if(!node->isLeaf())
	{
	  debug (15) << "__Parent " << node->toString() << "\n";
	  // this node is no longer a valid leaf node.
	  node->removeCopy();
	  original->leaves->erase(i);
	}
      else if(node->isDead()) 
	{
	  debug (15) << "__Dead " << node->toString() << "\n";
	  // this node is dead. This can't happen if everything is working correct.
	  cerr << "*BUG* Tree::commit() :" << node->toString() << " is dead. Tell Martin his parser sux!\n";
	  j = original->removeBranch(i);
	}
      else
	{
	  debug (15) << "__Live " << node->toString() << "\n";
	}
      i = j;
    }
  
  // go through this copy's leaf nodes
  i = begin();
  z = end();
  while(i != z)
    {
      j = i; ++j;
      RENode * node = *i;
      
      debug (15) << "__New " << node->toString() << "\n";
      // move the leaf to the original's leaf list.
      leaves->erase(i);
      original->leaves->push_front(node);
      
      i = j;
    }
}


string Tree::toString() const
{
  string s = "Tree[\n";
  for(RENodeI i = begin(), z = end();
      i!=z; ++i)
    {
      RENode * leaf = *i;
      s += "  " + leaf->toString() + "\n";
    }
  return s+"]Tree";
}
