#include "SRNode.hh"
#include "RENode.hh"
#include "Regex.hh"
#include "Symbol.hh"
#include "Token.hh"
#include "RETree.hh"
#include "debug.hh"

//unsigned int Rule::matchInvoked = 0;
unsigned int RegExpr::matchInvoked = 0;
//unsigned int State::cloneInvoked = 0;

/**
 * Literal regex (matches one token)
 * e.g.: TOKEN
 */
void LiteralExpr::match(Tree* tree) const
{
  ++RegExpr::matchInvoked;
  debug(2) << "LiteralExpr(" << toString() << ")::match " << tree->toString() << " {\n";
  if(tree->size()==0) return;
  
  // Iterate through all states in the tree.
  RENodeI ni=tree->begin();
  RENode * node;
  
  while(ni!=tree->end())
    {
      node = *ni;
      
      // Has this state reached the end of the input string?
      // If so, cut the branch and continue to the next state.
      if(node->getNext()->getNodes() == 0)
	{
	  ni = tree->removeBranch(ni);
	  continue;
	}
      
      vector<RENode*>* matches = (*ni)->getNext()->appendMatches(tree, ni, literal);
      for(vector<RENode*>::iterator
	    ii=matches->begin(),
	    iz=matches->end();
	  ii!=iz; ++ii)
	{
	  debug(15) << "MATCH: " << (*ii)->getMatchedNode()->getToken()->toString() << "\n";
	}
      
      if(matches->size() > 0)
	{
	  // Create a new state in the tree for each match.
	  ni = tree->addBranch(ni, literal, matches);
	}
      else
	{
	  // or cut the branch, if the match failed.
	  ni = tree->removeBranch(ni);
	  continue;
	}
      delete matches;
      
      ++ni;
    }
  debug(2) << "} LiteralExpr::match => " << tree->toString() << "\n";
}


/**
 * Repetition regex: * operator
 * e.g.: (EXPR *)
 * Note that the + operator is modelled as (EXPR (EXPR *) )
 */
void RepetitionExpr::match(Tree* tree) const
{
  ++RegExpr::matchInvoked;
  debug(2) << "RepetitionExpr(" << toString() << ")::match " << tree->toString() << " {\n";
  if(tree->size()==0) return;
  
  Tree * temp1 = new Tree(tree);
  Tree * temp2 = new Tree(tree);
  
  // match once;
  expr->match(temp1);
  
  int nMatches = temp1->size(); 
  
  // commit match-once-results
  temp1->commit(); // commit temp1 to tree (the originator)
  delete temp1;
  
  // tree now contains states that matched expr once.
  // temp2 contains the states that matched 0 times.
  
  // if there are any states left in tree, then match again.
  if(nMatches) match(tree);
  
  // commit match-zero-results
  temp2->commit();
  delete temp2;
  
  debug(2) << " } RepetitionExpr(" << toString() << ")::match => " << tree->toString() <<"\n";
}

/**
 * Optional regex: ?-operator
 * e.g.: (EXPR ?)
 */
void OptionalExpr::match(Tree* tree) const
{
  ++RegExpr::matchInvoked;
  debug(2) << "OptionalExpr(" << toString() << ")::match " << tree->toString() << " {\n";
  if(tree->size()==0) return;
  
  Tree * temp1 = new Tree(tree);
  Tree * temp2 = new Tree(tree);
  
  // Match once
  expr->match(temp1);
  
  // commit match-once-results
  temp1->commit(); // commit temp1 to tree (the originator)
  delete temp1;
  
  // tree now contains states that matched expr once.
  // temp2 contains the states that matched 0 times.
  
  // commit match-zero-results
  temp2->commit(); // commit temp2 to tree (the originator)
  delete temp2;
  
  debug(2) << "} OptionalExpr(" << toString() << ")::match => " << tree->toString() <<" \n";
}

/**
 * Alternation regex: |-operator
 * e.g.: (EXPR1 | EXPR2)
 */
void AlternationExpr::match(Tree* tree) const
{
  ++RegExpr::matchInvoked;
  debug(2) << "AlternationExpr(" << toString() << ")::match " << tree->toString() << " {\n";
  if(tree->size()==0) return;
  
  // get the two tree for alternative matches
  Tree * temp1 = new Tree(tree);
  Tree * temp2 = new Tree(tree);
  
  // Clear all 0-match states.
  tree->clear();
  
  // get matches from both alternatives
  expr1->match(temp1);
  expr2->match(temp2);
  
  // then join them together 
  temp1->commit(); // commit temp1 to tree (the originator)
  delete temp1;
  temp2->commit(); // commit temp2 to tree (the originator)
  delete temp2;
  
  debug(2) << " } AlternationExpr(" << toString() << ")::match => " << tree->toString() << "\n";
}


/**
 * Sequence regex: (implicit) .-operator
 * e.g.: (EXPR1 EXPR2)
 */
void SequenceExpr::match(Tree* tree) const
{
  ++RegExpr::matchInvoked;
  debug(2) << "SequenceExpr(" << toString() << ")::match " << tree->toString() << " {\n";
  if(tree->size()==0) return;
  
  // match first expression,
  expr1->match(tree);
  // and then the next expression.
  expr2->match(tree);
  
  debug(2) << " } SequenceExpr(" << toString() << ")::match => " << tree->toString() << "\n";
}

/**
 *
 */
void RegExpr::match(Tree* tree) const
{
  ++RegExpr::matchInvoked;
}

//
// Stringification
//

/**
 *
 */
string SequenceExpr::toString() const
{
  return "(" + expr1->toString() + " " + expr2->toString() + ")";
}

/**
 *
 */
string AlternationExpr::toString() const
{
  return "(" + expr1->toString() + "||" + expr2->toString() + ")";
}

/**
 *
 */
string LiteralExpr::toString() const
{
  return literal->toString();
}

/**
 *
 */
string RepetitionExpr::toString() const
{
  return ""+expr->toString()+"*";
}

/**
 *
 */
string OptionalExpr::toString() const
{
  return ""+expr->toString()+"?";
}

/**
 *
 */
string RegExpr::toString() const
{
  return "RegExpr";
}
