%{
#include <string>
#include <vector>
#include <errno.h>
#include <stdio.h>
#include "common.hh"
#include "Rule.hh"
#include "Regex.hh"
#include "Token.hh"
#include "Symbol.hh"
#include "debug.hh"
#include "compiler.hh"
#include "tokenizer.hh"
#include "SRNode.hh"
#include "RETree.hh"
  
  extern int errno;
  extern FILE * yyin;
  extern char yytext;
  extern int linie;
  int yylex();
  void yyerror(char *s)
    {
      cout << s << " in line " << linie << "\n";
    }
  int debugLevel;
  int reduce;
  int tree;
  int code;
  int parse;
  int experimental;
  int printComplete;
  int printIncompletable;
  int printAllStates;
  int tracking;
  int statistics;
  
  Rules * rules = new Rules();
  Synonyms * synonyms = new Synonyms();
  TokenString * input;
  TokenString * output;
  
%}

%union 
{
  string * str;
  TagField * field;
  FormAndFunc * ff;
  Symbol * sym;
  RegExpr * expr;
  vector<Rule*> * rulelist;
  StringSet * set;
}

%token <str> ID DOUBLEQOUTED SINGLEQOUTED REPLACE OR ENDRULES SUBST
%type <str> rewriter 
%type <expr> regex term factor
%type <set> set_of_tags id_list
%type <ff> form_and_function lit_list set_of_literals
%type <field> tag_field
%type <sym> symbol

%start grammar

%%

/**
 * Grammar description.
 */
grammar:
ruleset ENDRULES
{
  rout << "grammar -> ruleset ENDRULES\n";
  YYACCEPT;
}
|
ruleset
{
  rout << "grammar -> ruleset\n";
}
;

/**
 * List of rules.
 */
ruleset:
rule ruleset
{
  rout << "ruleset -> rule ruleset\n";
}
| /* epsilon */
{
  rout << "ruleset -> \n";
}
;

/**
 * Grouping or rewriting rule.
 * e.g.:
 *  X:pp = X[->H]:prp P<
 *  P:v <- {FMV,IMV}
 */
rule:
symbol '=' '+' regex ';'
{
  rout << "rule -> symbol =+ regex ;\n";
  Rule * rule = new Rule($1,$4,'+');
  rules->push_back(rule);
  debug(1) << rule->toString() << "\n";
}
| symbol '=' '!' regex ';'
{
  rout << "rule -> symbol =+ regex ;\n";
  Rule * rule = new Rule($1,$4,'!');
  rules->push_back(rule);
  debug(1) << rule->toString() << "\n";
}
| symbol '=' regex ';'
{
  rout << "rule -> symbol = regex ;\n";
  Rule * rule = new Rule($1,$3,' ');
  rules->push_back(rule);
  debug(1) << rule->toString() << "\n";
}
| symbol SUBST symbol ';'
{
  rout << "rule -> symbol <- symbol ;\n";
  Symbol * synonym = new Symbol
    (new TagField($3->funcs->tags,  $1->funcs->first()),
     new TagField($3->forms->tags,  $1->forms->first()),
     new TagField($3->sfuncs->tags, $1->sfuncs->first()),
     new TagField($3->sforms->tags, $1->sforms->first())
     );
  synonyms->push_back(synonym);
  debug(1) << synonym->toString() << "\n";
}
;

/**
 * Regular expression or union of two regexs.
 * e.g.:
 *  (N>+ X N<*) || (N>* X N<+)
 */
regex:
regex OR term
{
  rout << "regex -> regex || term\n";
  $$ = new AlternationExpr($1, $3);
}
| term
{
  rout << "regex -> term\n";
  $$ = $1;
}
;

/**
 * Term or concatenation of two regexs.
 * e.g.:
 *  N>+ X[->H]
 */
term:
term factor
{
  rout << "term -> term factor\n";
  $$ = new SequenceExpr($1, $2);
}
| factor
{
  rout << "term -> factor\n";
  $$ = $1;
}
;

/**
 * Factor. 
 * e.g.:
 *   ( >N+ X[->H] <N* )
 *   >N+
 *   X[->H]
 *   {IAUX,FAUX}[->DEP]
 */
factor:
symbol '.' set_of_literals
{
  rout << "factor -> symbol\n";
  if($1->isEmpty())
    {
      cerr << "An empty symbol (e.g. :_:) was apparently found in line " << linie << ". \nThe cause may be a missing ; after the rule before.\n";
      YYERROR;
    }
  $1->wordforms->tags = $3->form->tags;
  $1->baseforms->tags = $3->func->tags;
  $$ = new LiteralExpr($1);
}
| symbol
{
  rout << "factor -> symbol\n";
  if($1->isEmpty())
    {
      cerr << "An empty symbol (e.g. :_:) was apparently found in line " << linie << ". \nThe cause may be a missing ; after the rule before.\n";
      YYERROR;
    }
  $$ = new LiteralExpr($1);
}
| '(' regex ')'
{
  rout << "factor -> ( regex )\n";
  $$ = $2;
}
| factor '*'
{
  rout << "factor -> factor *\n";
  $$ = new RepetitionExpr($1);
}
| factor '+'
{
  rout << "factor -> factor +\n";
  RegExpr * newExpr = new RepetitionExpr($1);
  $$ = new SequenceExpr($1, newExpr);
}
| factor '?'
{
  rout << "factor -> factor ?\n";
  $$ = new OptionalExpr($1);
}
;

/**
 * Combined constituent _ sentence symbol, including rewriting rules
 * or a set of literals
 * e.g.:
 *   {FAUX,IAUX}[->DEP]_Y[->NIL]
 *   {"som","hvilken"}
 */
symbol:
form_and_function '_' form_and_function
{
  rout << "symbol -> form_and_function _ form_and_function\n";
  $$ = new Symbol($1->func, $1->form, $3->func, $3->form);
}
| form_and_function
{
  rout << "symbol -> form_and_function\n";
  $$ = new Symbol($1->func, $1->form);
}
| set_of_literals
{
  rout << "symbol -> literals\n";
  $$ = new Symbol(0,0,0,0, $1->form, $1->func);
}
;

/**
 * Combined function:form tags, including rewriting rules.
 * e.g.:
 *  X[->H]:adj
 *  N<
 *  {FAUX,IAUX}[->DEP]
 */
form_and_function:
tag_field ':' tag_field
{
  rout << "form_and_function -> tag-field : tag-field\n";
  $$ = new FormAndFunc($1, $3);
}
| tag_field
{
  rout << "form_and_function -> tag-field\n";
  $$ = new FormAndFunc($1);
}
;

tag_field:
set_of_tags rewriter
{
  rout << "tag_field -> set_of_tags rewriter\n";
  $$ = new TagField($1, *$2);
}
;

/**
 * set of 0 or more tags (id)
 * e.g.:
 *  X
 *  {n,np}
 */
set_of_tags:
'{' id_list '}'
{
  rout << "set_of_tags -> { id_list }\n";
  $$ = $2;
}
| ID
{
  rout << "set_of_tags -> ID(" << *$1 << ")\n";
  $$ = new StringSet();
  $$->insert(*$1);
}
|
{
  $$ = new StringSet();
}
;

/**
 * List of one or more identifiers.
 * e.g.:
 *  FAUX,IAUX
 *  X
 */
id_list:
ID ',' id_list
{
  rout << "id_list -> ID(" << *$1 << ") id_list\n";
  $$ = $3;
  $$->insert(*$1);
}
| ID
{
  rout << "id_list -> ID(" << *$1 << ")\n";
  $$ = new StringSet();
  $$->insert(*$1);
}
;



/**
 * set of 1 or more literals (strings)
 * e.g.:
 *  'der'
 *  {"som","hvilken"}
 */
set_of_literals:
'{' lit_list '}'
{
  rout << "literals -> { lit_list }\n";
  $$ = $2;
}
| DOUBLEQOUTED
{
  rout << "literals -> string(" << *$1 << ")\n";
  $$ = new FormAndFunc();
  $$->func->tags->insert(*$1);
}
| SINGLEQOUTED
{
  rout << "literals -> string(" << *$1 << ")\n";
  $$ = new FormAndFunc();
  $$->form->tags->insert(*$1);
}
;


/**
 * List of one or more literals ("- or '- encased strings)
 * '-encased strings are considered word forms.
 * "-encased strings are considered base forms.
 * e.g.:
 *  'did', 'does'
 *  "do"
 */
lit_list:
DOUBLEQOUTED ',' lit_list
{
  rout << "lit_list -> \"" << *$1 << "\" , lit_list\n";
  $$ = $3;
  $$->func->tags->insert(*$1);
}
| SINGLEQOUTED ',' lit_list
{
  rout << "lit_list -> \'" << *$1 << "\' , lit_list\n";
  $$ = $3;
  $$->form->tags->insert(*$1);
}
| DOUBLEQOUTED
{
  rout << "lit_list -> \"" << *$1 << "\"\n";
  $$ = new FormAndFunc();
  $$->func->tags->insert(*$1);
}
| SINGLEQOUTED
{
  rout << "lit_list -> \'" << *$1 << "\'\n";
  $$ = new FormAndFunc();
  $$->form->tags->insert(*$1);
}
;

/**
 * Rewriting rule.
 * e.g.:
 *  [->DEP]
 *  [->]
 */
rewriter:
'[' REPLACE ID ']'
{
  rout << "rewriter -> [-> ID(" << *$3 << ") ]\n";
  $$ = $3;
}
|
{
  $$ = new string(""); 
}
;

%%

void usage()
{
  cout << "Usage: cg2tree [options] grammar_file\n"
    "        Read cg output from stdin and try to compose\n"
    "        tree structures by following the rules\n"
    "        specified in grammar_file. All successfully\n"
    "        constructed trees are by default printed to\n"
    "        stdout in VISL format.\n\n"
    "Options:\n"
    //    "  -c          Generate source code for a static grammar.\n"
    "  -C          Print complete trees as soon as they are found.\n"
    "  -b          Don't attempt to build any trees.\n"
    "              Exit after printing the compiled rule set\n"
    "              and input tokens.\n"
    "  -dn         Turn on debugging output.\n"
    "              n specifies a debug level to turn on.\n"
    "              If n is unspecified or 0, level 1 is assumed.\n"
    "  -h          Print this help text and exit.\n"
    "  -I          Print incompletable states as soon as they are found.\n"
    "  -r          Print all reductions made when parsing the grammar file.\n"
    "  -s          Print some statistics.\n"
    "  -S          Print each state visited during compilation.\n"
    "  -t          Suppress output of VISL trees.\n"
    "  -T          For each complete tree, print the sequence\n"
    "              of rules used to build it.\n"
    "  -x          Use experimental algorithm.\n"
    ;
}

int main(int argc, char * argv[])
{
  parse = 1;
  code = 0;
  experimental = 0;
  debugLevel = 0;
  printComplete = 0;
  printIncompletable = 0;
  printAllStates = 0;
  reduce = 0;
  tree = 1;
  tracking = 0;
  statistics = 0;
  if(argc<=1)
    {
      usage();
      return 1;
    }
  for(int i=1; i<argc-1; i++)
    {
      if(argv[i][0] == '-')
	{
	  switch(argv[i][1])
	    {
	    case 'd': /* debug output on */
	      if(argv[i][2]==0)
		{
		  cerr << "-d : missing argument. Assuming level 1.\n";
		  debugLevel |= 1;
		  cerr << "Debug level = " << debugLevel << "\n";
		}
	      else
		{
		  debugLevel |= argv[i][2] - '0';
		  cerr << "Debug level = " << debugLevel << "\n";
		}
	      break;
	    case 'c': /* generate compiler-source */
	      cerr << "warning: code generation not implemented (-c)\n";
	      code = 1;
	      break;
	    case 'r': /* reductions output on */
	      reduce = 1;
	      break;
	    case 'b': /* don't build. quit after compiling rules and tokens */
	      parse = 0;
	      debugLevel |= 1;
	      break;
	    case 'C': /* list all complete trees as they are found */
	      printComplete = 1;
	      break;
	    case 'h': /* print help text */
	      usage();
	      return 0;
	    case 'I': /* list all incomplete trees as they are found */
	      printIncompletable = 1;
	      break;
	    case 's': /* print "profiling" statistics */
	      statistics = 1;
	      break;
	    case 'S': /* print all states visited. */
	      printAllStates = 1;
	      break;
	    case 'T': /* track state history */
	      tracking = 1;
	      break;
	    case 't': /* don't output VISL trees. */
	      tree = 0;
	      break;
	    case 'x': /* use experimental algorithm. */
	      experimental = 1;
	      break;
	    default:
	      cout << "Unknown option " << argv[i] << "\n";
	      usage();
	      return 1;
	    }
	} 
      else 
	{
	  usage();
	  return 1;
	}
    }
  // set input source of lexer.
  yyin = fopen(argv[argc-1], "r" );
  if(yyin==NULL)
    {
      cout << "Could not open file " << argv[argc-1] <<"\n";
      return 1;
    }
  debug(1) << "Compiled Rules:\n";
  if(!yyparse())
    {
      Token * t = 0;
      Token * prev = 0;
      Token * last = 0;
      output= new TokenString();
      debug(1) << "\nInput Tokens:\n";
      while((t = readToken(prev)))
	{
	  debug(1) << t->toString() << "\n";
	  prev = t;
	}
      if(parse)
	{
	  doParse(prev);
	}
      if(statistics)
	{
	  cout << rules->size() << " rules.\n";
	  cout << synonyms->size() << " synonym definitions.\n";
	  cout << input->size() << " input tokens.\n";
	  cout << output->size() << " output trees.\n";
	  cout << "Tokens created: "<<Token::constructorInvoked<<"\n";
	  cout << "Token::appendMatches invoked "<<Token::appendMatchesInvoked<<" times.\n";
	  cout << "Rule::match invoked "<<Rule::matchInvoked<<" times.\n";
	  cout << "RegExpr::match invoked "<<RegExpr::matchInvoked<<" times.\n";
	}
    }
  else
    {
      cout << "Failed to parse rules!\n";
    }
}
